# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import os
import base64
import uuid
# import webbrowser
import urllib.parse
import urllib.request


from ._debug_utils import debug_print
from .ipython_api import display, HTML
from .ipython_api import IPythonAPI
from .my_utils import adjust_path, adjust_path_to_uri, json_dumps, single_quote
from .constants import Constants
from .os_dependent_api import OsDependentAPI


try:
    from pygments import highlight
    from pygments.lexers.data import JsonLexer
    from pygments.formatters.terminal import TerminalFormatter
except:
    highlight = None
    JsonLexer = None
    TerminalFormatter = None


class FormattedJsonDict(dict):

    def __init__(self, item, *args, **kwargs):
        super(FormattedJsonDict, self).__init__()
        self.row = None
        if type(item).__name__ == 'KqlRow':
            self.row = item
            item = item.row
        self.update(item)

        _dict = self
        if len(item) == 1 and isinstance(self.get(" "), list):
            _dict = self.get(" ")
        formatted_json = json_dumps(_dict, indent=4, sort_keys=True)
        self.colorful_json = formatted_json

        if highlight and JsonLexer and TerminalFormatter:
            try:
                self.colorful_json = highlight(formatted_json.encode("UTF-8"), JsonLexer(), TerminalFormatter())
            except:
                pass

    def get(self, key, default=None):
        value = super(FormattedJsonDict, self).get(key, default)
        return _getitem_FormattedJson(value)


    def __getitem__(self, key):
        if self.row is not None:
            value = self.row.__getitem__(key)
        else:
            value = self.get(key)
        return _getitem_FormattedJson(value)


    def __repr__(self):
        return self.colorful_json


class FormattedJsonList(list):

    def __init__(self, item, key, *args, **kwargs):
        super(FormattedJsonList, self).__init__()
        self.extend(item)
        self.item = item
        self.key = key or " "
        # formatted_json = json_dumps(self.item, indent=4, sort_keys=True)
        # self.colorful_json = highlight(formatted_json.encode("UTF-8"), JsonLexer(), TerminalFormatter())


    def __getitem__(self, key):
        value = super(FormattedJsonList, self).__getitem__(key)
        return _getitem_FormattedJson(value)


    def __repr__(self):
        if len(self.item) > 0 and type(self.item[0]).__name__ == 'KqlRow':
            _list = [kql_row.row for kql_row in self.item]
        else:
            _list = self.item
        return FormattedJsonDict({self.key: _list}).__repr__()
        # return self.colorful_json


def _getitem_FormattedJson(item, key=None):
    if isinstance(item, list):
        return FormattedJsonList(item, key)
    elif isinstance(item, dict) or type(item).__name__ == "KqlRow":
        return FormattedJsonDict(item)
    else:
        return item


class JSONDict(dict):

    def __init__(self, item, *args, **kwargs):
        super(JSONDict, self).__init__()
        self.item = item
        self.row = None
        if type(item).__name__ == "KqlRow":
            self.row = item
            item = item.row
        self.update(item)


    def get(self, key, default=None):
        value = super(JSONDict, self).get(key, default)
        return _getitem_JSON(value, key or default)


    def __getitem__(self, key):
        if self.row is not None:
            value = self.row.__getitem__(key)
        else:
            value = self.get(key)
        return _getitem_JSON(value, key)

    def _repr_json_(self):
        return self


    def __repr__(self):
        return FormattedJsonDict(self.item).__repr__()


class JSONList(list):

    def __init__(self, item, key, *args, **kwargs):
        super(JSONList, self).__init__()
        self.item = item
        self.extend(item)
        self.key = key or " "


    def __getitem__(self, key):
        value = super(JSONList, self).__getitem__(key)
        return _getitem_JSON(value, key)


    def _repr_json_(self):
        if len(self) > 0 and type(self.item[0]).__name__ == "KqlRow":
            _list = [kql_row.row for kql_row in self.item]
        else:
            _list = self.item
        return JSONDict({self.key: _list})._repr_json_()

    def __repr__(self):
        return FormattedJsonList(self.item, self.key).__repr__()


def _getitem_JSON(item, key=None):
    if isinstance(item, list):
        return JSONList(item, key)
    elif isinstance(item, dict) or type(item).__name__ == "KqlRow":
        return JSONDict(item)
    else:
        return item


class Display(object):
    """
    """

    success_style = {"color": "#417e42", "background-color": "#dff0d8", "border-color": "#d7e9c6"}
    danger_style = {"color": "#b94a48", "background-color": "#f2dede", "border-color": "#eed3d7"}
    info_style = {"color": "#3a87ad", "background-color": "#d9edf7", "border-color": "#bce9f1"}
    warning_style = {"color": "#8a6d3b", "background-color": "#fcf8e3", "border-color": "#faebcc"}
    ref_style = {
        "padding": "2px 6px 2px 6px",
        "color": "#333333",
        "background-color": "#EEEEEE",
        "border-top": "1px solid #CCCCCC", 
        "border-right": "1px solid #333333", 
        "border-bottom": "1px solid #333333", 
        "border-left": "1px solid #CCCCCC"
    }

    showfiles_url_base_path = None
    showfiles_file_base_path = None
    showfiles_folder_name = None


    @staticmethod
    def show_html(html_str, display_handler_name=None, **options):
        Display.show_html_obj(HTML(html_str), display_handler_name=display_handler_name, **options)


    @staticmethod
    def show_html_obj(html_obj, display_handler_name=None, **options) -> object:
        if display_handler_name is None:
            display(html_obj)
        else:
            dh = options["display_handlers"].get(display_handler_name)
            if dh is not None:
                dh.update(html_obj)
            else:
                options["display_handlers"][display_handler_name] = display(html_obj, display_id=options.get("display_id"))
        

    @staticmethod
    def show(content, display_handler_name=None, **options):
        if isinstance(content, str) and len(content) > 0:
            if options is not None and options.get("popup_window", False):
                file_name = Display._get_name(**options)
                file_ext = None
                if options.get("notebook_app") in ["azuredatastudiosaw"]:
                    file_ext = "txt"
                file_path = Display._html_to_file_path(content, file_name, file_ext=file_ext, **options)
                Display.show_window(file_name, file_path, display_handler_name=display_handler_name, **options)
            else:
                Display.show_html(content, display_handler_name=display_handler_name, **options)
        elif display_handler_name is None:
            display(content)
        else:
            dh = options["display_handlers"].get(display_handler_name)
            if dh is not None:
                dh.update(content)
            else:
                options["display_handlers"][display_handler_name] = display(content, display_id=options.get("display_id"))


    @staticmethod
    def get_show_deeplink_webbrowser_html_obj(window_name, deep_link_url:str, close_window_timeout_in_secs: int, options=None):
        options = options or {}
        close_itself_timeout_in_secs = 0
        html_str = Display._get_Launch_page_html(window_name, deep_link_url, close_window_timeout_in_secs, close_itself_timeout_in_secs, False, options=options)
        file_name = Display._get_name()
        file_path = Display._html_to_file_path(html_str, file_name, **options)
        url = Display._get_file_path_url(file_path, options=options)
        return url


    @staticmethod
    def get_show_deeplink_html_obj(window_name, deep_link_url:str, close_window_timeout_in_secs: int, options=None):
        options = options or {}
        if options.get("kernel_location") == "local":
            url = Display.get_show_deeplink_webbrowser_html_obj(window_name, deep_link_url, close_window_timeout_in_secs, options=options)
            OsDependentAPI.webbrowser_open(url)
            Display.showInfoMessage(f"opened popup window: {window_name}, see your browser")
            return None
        else:
            close_itself_timeout_in_secs = None
            html_str = Display._get_Launch_page_html(window_name, deep_link_url, close_window_timeout_in_secs, close_itself_timeout_in_secs, False, options=options)
            return HTML(html_str)


    @staticmethod
    def get_show_window_html_obj(window_name, file_path, button_text=None, onclick_visibility=None, isText:bool=None, palette:dict=None, before_text=None, after_text=None, close_window_timeout_in_secs=None, content=None, options=None):
        options = options or {}
        html_str = None
        mode = options.get("popup_interaction", "auto")
        if mode == "auto":
            if options.get("notebook_app") in ["ipython"] and options.get("test_notebook_app") in ["none", "ipython"]: 
                mode = "webbrowser_open_at_kernel"
            elif options.get("notebook_app") in ["visualstudiocode", "azuredatastudiosaw", "azuredatastudio"] and options.get("test_notebook_app") in ["none", "visualstudiocode", "azuredatastudio", "azuredatastudiosaw"]: 
                mode = "reference"
            elif options.get("notebook_app") in ["nteract"] and options.get("temp_files_server_address") is None:
                mode = "reference"
            else:
                mode = "button"
        if mode == "webbrowser_open_at_kernel":
            url = Display._get_file_path_url(file_path, options=options)
            # url = urllib.parse.quote(url)
            OsDependentAPI.webbrowser_open(url)
            html_str = Display._getInfoMessageHtmlStr(f"opened popup window: '{window_name}', see it in your browser", **options)

        elif mode in ["button", "memory_button"]:
            html_str = Display._get_window_html(window_name, file_path, button_text, onclick_visibility, isText=isText, palette=palette, before_text=before_text, after_text=after_text, close_window_timeout_in_secs=close_window_timeout_in_secs, content=content, options=options)

        elif mode in ["reference", "reference_popup"]:
            html_str = Display._get_window_ref_html(mode == "reference_popup", window_name, file_path, button_text, isText=isText, palette=palette, before_text=before_text, after_text=after_text, close_window_timeout_in_secs=close_window_timeout_in_secs, options=options)

        return HTML(html_str) if html_str is not None else None


    ############################################################################
    #
    ############################################################################
    @staticmethod
    def _get_file_path_url(file_path, options=None):
        options = options or {}
        url = None
        if file_path.startswith("http"):
            url = file_path

        elif options.get("temp_files_server_address") is not None and Display.showfiles_url_base_path.startswith("http"):
            url = f'{Display.showfiles_url_base_path}/{adjust_path_to_uri(file_path)}?kernelid={options.get("kernel_id")}'

        else:
            url = Display._get_file_path_file_url(file_path, options=options)

        return url

    @staticmethod
    def _get_file_path_file_url(file_path, options=None):
        options = options or {}
        url = None
        path_uri = adjust_path_to_uri(f"{Display.showfiles_file_base_path}/{file_path}")
        url = (f"file:///{path_uri}")

        return url


    @staticmethod
    def _get_window_ref_html(isPopupMode, window_name, file_path, ref_text=None, isText=None, palette=None, ref_palette=None, before_text=None, after_text=None, close_window_timeout_in_secs=None, options=None):
        options = options or {}
        url = Display._get_file_path_url(file_path, options=options)
        popup_window_name = window_name

        if isPopupMode or options.get("temp_files_server_address") is not None:
            close_window_timeout_in_secs = close_window_timeout_in_secs or 60  # five minutes
            popup_window_name = f"popup_{window_name}"
            popup_html = Display._get_popup_window_html(url, window_name, close_window_timeout_in_secs, options=options)
            popup_file_name = f"popup_{file_path.split('/')[-1].split('.')[0]}"
            popup_file_path = Display._html_to_file_path(popup_html, popup_file_name, **options)
            url = Display._get_file_path_url(popup_file_path, options=options)

        if options.get("temp_files_server_address") is not None:
            indirect_url = f'{options.get("temp_files_server_address")}/webbrowser?url={urllib.parse.quote(url)}&kernelid={options.get("kernel_id")}'
            url = indirect_url
        ref_text = ref_text or "popup window"
        
        style_div = f"padding: 10px; color: {palette.get('color')}; background-color: {palette.get('background-color')}; border-color: {palette.get('border-color')}" if palette else ""
        ref_palette = ref_palette or Display.ref_style
        style_ref = f"padding: 2px 6px 2px 6px; color: {ref_palette['color']}; background-color: {ref_palette['background-color']}; border-top: {ref_palette['border-top']}; border-right: {ref_palette['border-right']}; border-bottom: {ref_palette['border-bottom']}; border-left: {ref_palette['border-left']}"

        before_text = f"{before_text.replace(' ', '&nbsp;')}&nbsp;" if before_text is not None else ''
        after_text = f"&nbsp;{after_text.replace(' ', '&nbsp;')}" if after_text is not None else ''
        html_str = (
            f"""<!DOCTYPE html>
            <html><body>
            <div style='{style_div}'>
            {before_text}<a href='{url}' style='{style_ref}' target='{popup_window_name}'>{ref_text}</a>{after_text}
            </div>
            </body></html>"""
        )

        return html_str


    @staticmethod
    def show_window(window_name, file_path, button_text=None, onclick_visibility=None, isText:bool=None, palette:dict=None, before_text=None, after_text=None, display_handler_name=None, close_window_timeout_in_secs=None, content=None, **options):
        html_obj = Display.get_show_window_html_obj(window_name, file_path, button_text=button_text, onclick_visibility=onclick_visibility, isText=isText, palette=palette, before_text=before_text, after_text=after_text, close_window_timeout_in_secs=close_window_timeout_in_secs, content=content, options=options)
        if html_obj is not None:
            Display.show_html_obj(html_obj, display_handler_name=display_handler_name, **options)


    @staticmethod
    def to_json_styled_class(item, style=None, options=None):
        options = options or {}
        json_display_style = style or options.get("json_display")
        if json_display_style != "raw" and (isinstance(item, list) or isinstance(item, dict) or type(item).__name__ == "KqlRow"):
            if json_display_style == "formatted":
                return _getitem_FormattedJson(item)
            else:
                return _getitem_JSON(item)
        else:
            return item


    @staticmethod
    def _html_to_file_path(html_str, file_name, file_ext: str=None, **options):
        file_path = f"{Display.showfiles_folder_name}/{file_name}.{file_ext or 'html'}"
        full_file_name = adjust_path(f"{Display.showfiles_file_base_path}/{file_path}")
        text_file = open(full_file_name, "wb")
        text_file.write(bytes(html_str, 'utf-8'))
        text_file.close()
        # ipython will delete file at shutdown or by restart
        IPythonAPI.try_add_to_ipython_tempfiles(full_file_name)
        return file_path


    @staticmethod
    def _get_name(**kwargs):
        if kwargs is not None and isinstance(kwargs.get("file_name"), str) and len(kwargs.get("file_name")) > 0:
            name = kwargs.get("file_name")
        else:
            name = uuid.uuid4().hex
        return name


    @staticmethod
    def _get_popup_window_html(url, window_name, close_window_timeout_in_secs, options=None):
        options = options or {}
        window_params = "fullscreen=no,directories=no,location=no,menubar=no,resizable=yes,scrollbars=yes,status=no,titlebar=no,toolbar=no,"
        window_name = window_name.replace(".", "_").replace("-", "_").replace("/", "_").replace(":", "_").replace(" ", "_")

        html_str = (
            f"""<!DOCTYPE html>
            <html><body>
            <script>
            function kql_MagicPopupWindowFunction(url, window_name, window_params, close_window_timeout_in_secs) {{
                window.focus();
                var w = screen.width / 2;
                var h = screen.height / 2;
                var params = 'width=' + w + ',height=' + h;
                var new_window = window.open(url, window_name, window_params + params);
                // close self window after new window is open
                if (new_window != null) {{
                    window.close();
                    close_window_timeout_in_secs = 1;
                }}
                setTimeout(function(){{ window.close(); }}, close_window_timeout_in_secs * 1000);               
            }}

            kql_MagicPopupWindowFunction('{url}','{window_name}','{window_params}',{str(close_window_timeout_in_secs)});
            window.focus();
            </script>
            <p>This page popups <b>'{window_name}'</b> window. It will close itself in few minutes.
            <br><br><br>If <b>'{window_name}'</b> window doesn't popup, popups are probably blocked on this
            page.<br>To enable the popup, you should modify your browser settings to allow popups on pages from this host: http://127.0.0.1:5000.
            To open popup manually press <a href='{url}'>here</a>
            <br><br><br><br><br><br><b>Note:</b> You can disable the popups in your notebook, by setting popup_interaction option 
            to 'reference' (will open in a tab) or 'webbrowser_open_at_kernel' (will auto open in a tab on the python kernel host)
            or 'button'. Some modes are not supported by some jupyter based implementations (try and find out).
            <br> To set the default mode in your notebook, run: %config Kqlmagic.popup_interaction={{mode}}
            <br> To set the mode for the current kql magic execution, add the option -pi '{{mode}}'
            </p>
            </body></html>"""
        )
        return html_str

    @staticmethod
    def _get_window_html(window_name, file_path_or_data, button_text=None, onclick_visibility=None, isText=None, palette=None, before_text=None, after_text=None, close_window_timeout_in_secs=None, close_itself_timeout_in_secs=None, content=None, options=None):
        options = options or {}
        # if isText is True, file_path_or_data is the text
        host_or_text = 'text' if isText else (options.get("notebook_service_address") or "")
        if host_or_text != 'text' and options.get("popup_interaction") == "memory_button":
            if file_path_or_data.startswith("http"):
                with urllib.request.urlopen(file_path_or_data) as bytes_reader:
                    file_path_or_data = bytes_reader.read().decode('utf-8')
            else:
                full_file_name = adjust_path(f"{Display.showfiles_file_base_path}/{file_path_or_data}")
                with open(full_file_name, 'r') as str_reader:
                    # Read the entire file
                    file_path_or_data = str_reader.read()
            file_path_or_data =  base64.b64encode(file_path_or_data.encode('utf-8')).decode('utf-8')
            isText = True
            host_or_text = 'body'
        onclick_visibility = "visible" if onclick_visibility == "visible" else "hidden"
        button_text = button_text or "popup window"
        window_name = window_name.replace(".", "_").replace("-", "_").replace("/", "_").replace(":", "_").replace(" ", "_")
        if window_name[0] in "0123456789":
            window_name = f"w_{window_name}"
        window_params = "fullscreen=no,directories=no,location=no,menubar=no,resizable=yes,scrollbars=yes,status=no,titlebar=no,toolbar=no,"

        style = f"padding: 10px; color: {palette['color']}; background-color: {palette['background-color']}; border-color: {palette['border-color']}" if palette else ""
        before_text = before_text or ''
        after_text = after_text or ''

        close_window_timeout_in_secs = close_window_timeout_in_secs if close_window_timeout_in_secs is not None else -1
        if close_itself_timeout_in_secs is None:
            close_itself_timeout_in_secs =  -1
        elif close_window_timeout_in_secs >= 0:
            close_itself_timeout_in_secs = max(close_itself_timeout_in_secs - close_window_timeout_in_secs, 0)

        if not isText and Display.showfiles_url_base_path.startswith("http"):
            file_path_or_data = Display._get_file_path_url(file_path_or_data, options=options)

            if options.get("temp_files_server_address") is not None and content == "schema" and options.get("notebook_app") in ["nteract"]:
                indirect_url = f'{options.get("temp_files_server_address")}/webbrowser?url={urllib.parse.quote(file_path_or_data)}&kernelid={options.get("kernel_id")}'
                file_path_or_data = indirect_url

            
        html_str = (
            f"""<!DOCTYPE html>
            <html><body>
            <div style='{style}'>
            {before_text}
            <button onclick="this.style.visibility='{onclick_visibility}';
            kql_MagicLaunchWindowFunction({single_quote(file_path_or_data)}, {single_quote(window_params)}, {single_quote(window_name)}, {single_quote(host_or_text)});
            kql_MagicCloseWindow(kql_Magic_{window_name}, {str(close_window_timeout_in_secs)}, {str(close_itself_timeout_in_secs)});">
            {button_text}</button>{after_text}
            </div>

            <script>
            var kql_Magic_{window_name} = null;


            function kql_MagicCloseWindow(window_obj, obj_secs, itself_secs) {{
                if (obj_secs >= 0) {{
                    _timeout = setTimeout(function(){{
                        window_obj.close();
                        if (itself_secs >= 0) {{
                            __timeout = setTimeout(function(){{window.close();}}, itself_secs * 1000);
                        }}
                    }}, obj_secs * 1000);
                }} else if (itself_secs >= 0) {{
                    _timeout = setTimeout(function(){{window.close();}}, itself_secs * 1000);
                }}
            }}

            function kql_MagicLaunchWindowFunction(file_path_or_data, window_params, window_name, host_or_text) {{
                var url;
                const baseURI = String(window.location);
                if (host_or_text == 'text' || host_or_text == 'body') {{
                    url = ''
                }} else if (file_path_or_data.startsWith('http')) {{
                    url = file_path_or_data;
                }} else if (host_or_text.endsWith('.azureml.ms') || host_or_text.endsWith('.azureml.net')) {{
                    let azuremlBaseURI = String(window.document.baseURI);
                    let start = azuremlBaseURI.search('activeFilePath=');
                    if (start > 0) {{
                        start += 'activeFilePath='.length;
                        let end = azuremlBaseURI.substring(start).search('&');
                        if (end < 0) {{
                            end = undefined;
                        }}
                        let parts = azuremlBaseURI.substring(start, end).split('/');
                        parts.pop();
                        url = host_or_text + '/tree/' + parts.join('/') + '/' + file_path_or_data;
                    }} else {{
                        var parts = baseURI.split('/');
                        parts.pop();
                        url = parts.join('/') + '/' + file_path_or_data;
                    }}
                }} else {{
                    var base_url = '';

                    // check if azure notebook
                    var azure_host = (host_or_text == null || host_or_text.length == 0) ? 'https://notebooks.azure.com' : host_or_text;
                    var start = azure_host.search('//');
                    var azure_host_suffix = '.' + azure_host.substring(start+2);

                    var end = baseURI.search(azure_host_suffix);
                    start = baseURI.search('//');
                    if (start > 0 && end > 0) {{
                        // # azure notebook environment, assume template: https://library-user.notebooks.azure.com
                        var parts = baseURI.substring(start+2, end).split('-');
                        if (parts.length == 2) {{
                            var library = parts[0];
                            var user = parts[1];
                            base_url = azure_host + '/api/user/' +user+ '/library/' +library+ '/html/';
                        }}
                    }}

                    // check if local jupyter lab
                    if (base_url.length == 0) {{
                        var configDataScipt  = document.getElementById('jupyter-config-data');
                        if (configDataScipt != null) {{
                            var jupyterConfigData = JSON.parse(configDataScipt.textContent);
                            if (jupyterConfigData['appName'] == 'JupyterLab' && jupyterConfigData['serverRoot'] != null &&  jupyterConfigData['treeUrl'] != null) {{
                                var basePath = {single_quote(Display.showfiles_file_base_path)} + '/';
                                if (basePath.startsWith(jupyterConfigData['serverRoot'])) {{
                                    base_url = '/files/' + basePath.substring(jupyterConfigData['serverRoot'].length+1);
                                }}
                            }}
                        }}
                    }}

                    // works for local jupyter notebook
                    if (base_url.length == 0) {{

                        var parts = baseURI.split('/');
                        parts.pop(); // remove notebook name segment
                        base_url = parts.join('/') + '/';
                    }}
                    url = base_url + file_path_or_data;
                }}

                window.focus();
                var w = screen.width / 2;
                var h = screen.height / 2;
                params = 'width='+w+',height='+h;
                // kql_Magic + window_name should be a global variable 
                window_obj = window.open(url, window_name, window_params + params);
                if (url == '') {{
                    let decodedData = atob(file_path_or_data);
                    if (host_or_text == 'text') {{
                        var el = window_obj.document.createElement('p');
                        window_obj.document.body.overflow = 'auto';
                        el.style.top = 0;
                        el.style.left = 0;
                        el.innerHTML = decodedData;
                        window_obj.document.body.appendChild(el);
                    }} else {{
                        window_obj.document.body.innerHTML = decodedData;
                    }}
                }}
                kql_Magic_{window_name} = window_obj;
            }}
            </script>

            </body></html>"""
        )
        return html_str

    @staticmethod
    def _get_Launch_page_html(window_name, file_path_or_data, close_window_timeout_in_secs, close_itself_timeout_in_secs, isText, options=None):
        options = options or {}
        # if isText is True, file_path_or_data is the text
        host_or_text = 'text' if isText else (options.get("notebook_service_address") or "")
        if host_or_text != 'text' and options.get("popup_interaction") == "memory_button":
            if file_path_or_data.startswith("http"):
                with urllib.request.urlopen(file_path_or_data) as bytes_reader:
                    file_path_or_data = bytes_reader.read().decode('utf-8')
            else:
                full_file_name = adjust_path(f"{Display.showfiles_file_base_path}/{file_path_or_data}")
                with open(full_file_name, 'r') as str_reader:
                    # Read the entire file
                    file_path_or_data = str_reader.read()
            file_path_or_data =  base64.b64encode(file_path_or_data.encode('utf-8')).decode('utf-8')
            isText = True
            host_or_text = 'body'
        window_name = window_name.replace(".", "_").replace("-", "_").replace("/", "_").replace(":", "_").replace(" ", "_")
        if window_name[0] in "0123456789":
            window_name = f"w_{window_name}"
        # negative means not to colose window
        close_window_timeout_in_secs = close_window_timeout_in_secs if close_window_timeout_in_secs is not None else -1
        if close_itself_timeout_in_secs is None:
            close_itself_timeout_in_secs =  -1
        elif close_window_timeout_in_secs >= 0:
            close_itself_timeout_in_secs = max(close_itself_timeout_in_secs - close_window_timeout_in_secs, 0)

        window_params = "fullscreen=no,directories=no,location=no,menubar=no,resizable=yes,scrollbars=yes,status=no,titlebar=no,toolbar=no,"

        html_str = (
            f"""<!DOCTYPE html>
            <html><body>
            <script>
            var kql_Magic_{window_name} = null;
            var kql_Magic_{window_name}_timeout = null;

            function kql_MagicCloseWindow(window_obj, obj_secs, itself_secs) {{
                if (obj_secs >= 0) {{
                    _timeout = setTimeout(function(){{
                        window_obj.close();
                        if (itself_secs >= 0) {{
                            __timeout = setTimeout(function(){{window.close();}}, itself_secs * 1000);
                        }}
                    }}, obj_secs * 1000);
                }} else if (itself_secs >= 0) {{
                    _timeout = setTimeout(function(){{window.close();}}, itself_secs * 1000);
                }}
            }}

            function kql_MagicLaunchWindowFunction(file_path_or_data, window_params, window_name, host_or_text) {{
                var url;
                const baseURI = String(window.location);
                if (host_or_text == 'text' || host_or_text == 'body') {{
                    url = ''
                }} else if (file_path_or_data.startsWith('http')) {{
                    url = file_path_or_data;
                }} else if (host_or_text.endsWith('.azureml.ms') || host_or_text.endsWith('.azureml.net')) {{
                    let azuremlBaseURI = String(window.document.baseURI);
                    let start = azuremlBaseURI.search('activeFilePath=');
                    if (start > 0) {{
                        start += 'activeFilePath='.length;
                        let end = azuremlBaseURI.substring(start).search('&');
                        if (end < 0) {{
                            end = undefined;
                        }}
                        let parts = azuremlBaseURI.substring(start, end).split('/');
                        parts.pop();
                        url = host_or_text + '/tree/' + parts.join('/') + '/' + file_path_or_data;
                    }} else {{
                        var parts = baseURI.split('/');
                        parts.pop();
                        url = parts.join('/') + '/' + file_path_or_data;
                    }}
                }} else {{
                    var base_url = '';

                    // check if azure notebook
                    var azure_host = (host_or_text == null || host_or_text.length == 0) ? 'https://notebooks.azure.com' : host_or_text;
                    var start = azure_host.search('//');
                    var azure_host_suffix = '.' + azure_host.substring(start+2);

                    var end = baseURI.search(azure_host_suffix);
                    start = baseURI.search('//');
                    if (start > 0 && end > 0) {{
                        var parts = baseURI.substring(start+2, end).split('-');
                        if (parts.length == 2) {{
                            var library = parts[0];
                            var user = parts[1];
                            base_url = azure_host + '/api/user/' +user+ '/library/' +library+ '/html/';
                        }}
                    }}

                    // check if local jupyter lab
                    if (base_url.length == 0) {{
                        var configDataScipt  = document.getElementById('jupyter-config-data');
                        if (configDataScipt != null) {{
                            var jupyterConfigData = JSON.parse(configDataScipt.textContent);
                            if (jupyterConfigData['appName'] == 'JupyterLab' && jupyterConfigData['serverRoot'] != null &&  jupyterConfigData['treeUrl'] != null) {{
                                var basePath = {single_quote(Display.showfiles_file_base_path)} + '/';
                                if (basePath.startsWith(jupyterConfigData['serverRoot'])) {{
                                    base_url = '/files/' + basePath.substring(jupyterConfigData['serverRoot'].length+1);
                                }}
                            }}
                        }}
                    }}

                    // works for local jupyter
                    if (base_url.length == 0) {{
                        var parts = baseURI.split('/');
                        parts.pop();
                        base_url = parts.join('/') + '/';
                    }}
                    url = base_url + file_path_or_data;
                }}

                window.focus();
                var w = screen.width / 2;
                var h = screen.height / 2;
                params = 'width='+w+',height='+h;
                // kql_Magic + window_name should be a global variable 
                window_obj = window.open(url, window_name, window_params + params);
                if (url == '') {{
                    let decodedData = atob(file_path_or_data);
                    if (host_or_text == 'text') {{
                        var el = window_obj.document.createElement('p');
                        window_obj.document.body.overflow = 'auto';
                        el.style.top = 0;
                        el.style.left = 0;
                        el.innerHTML = decodedData;
                        window_obj.document.body.appendChild(el);
                    }} else {{
                        window_obj.document.body.innerHTML = decodedData;
                    }}
                }}
                kql_Magic_{window_name} = window_obj;
            }}

            kql_MagicLaunchWindowFunction({single_quote(file_path_or_data)}, {single_quote(window_params)}, {single_quote(window_name)}, {single_quote(host_or_text)});

            kql_MagicCloseWindow(kql_Magic_{window_name}, {str(close_window_timeout_in_secs)}, {str(close_itself_timeout_in_secs)});

            </script>
            </body></html>"""
        )
        return html_str

    @staticmethod
    def toHtml(**kwargs):
        title = '' if kwargs.get('title') is None else f"<title>{Constants.MAGIC_PACKAGE_NAME} - {kwargs.get('title')}</title>"
        return f"""<html>
        <head>
            {kwargs.get('head', '')}
            {title}
        </head>
        <body>
            {kwargs.get('body', '')}
        </body>
        </html>"""


    @staticmethod
    def _getMessageHtml(msg, palette):
        "get query information in as an HTML string"

        style_options = {}
        if isinstance(msg, dict):
            if isinstance(msg.get("style_options"), dict):
                style_options = msg.get("style_options")
                msg = msg.get("message")
        
        if isinstance(msg, list):
            msg_str = "<br>".join(msg)
        elif isinstance(msg, dict):
            msg_str = "<br>".join([f"{k}: {v}" for k, v in msg.items()])
        elif isinstance(msg, str):
            msg_str = msg
        else:
            msg_str = f"{msg or ''}"
        if len(msg_str) > 0:          
            style = {
                "padding": "10px",
                "color": palette["color"],
                "background-color": palette["background-color"],
                "border-color": palette["border-color"],
                **style_options
            }
            style_str = "".join([f"; {k}: {v}" for k,v in style.items()])[1:]
            msg_str = msg_str.replace('"', "&quot;").replace("'", "&apos;").replace("\n", "<br>").replace(" ", "&nbsp;")
            body = f"<div><p style='{style_str}'>{msg_str}</p></div>"
        else:
            body = ""
        return {"body": body}


    @staticmethod
    def getSuccessMessageHtml(msg):
        return Display._getMessageHtml(msg, Display.success_style)


    @staticmethod
    def getInfoMessageHtml(msg):
        return Display._getMessageHtml(msg, Display.info_style)


    @staticmethod
    def getWarningMessageHtml(msg):
        return Display._getMessageHtml(msg, Display.warning_style)


    @staticmethod
    def getDangerMessageHtml(msg):
        return Display._getMessageHtml(msg, Display.danger_style)

    @staticmethod
    def _getMessageHtmlStr(html_msg, display_handler_name=None, **options):
        html_str = None
        if html_msg is None or len(html_msg["body"]) == 0:
            if display_handler_name is not None:
                if options.get("display_id", False):
                    html_str = Display.toHtml(**html_msg)
                # dh = options["display_handlers"].get(display_handler_name)
                # if dh is not None:
                #     html_str = Display.toHtml(**html_msg)
        else:
            html_str = Display.toHtml(**html_msg)

        return html_str


    @staticmethod
    def _showMessage(html_msg, display_handler_name=None, **options):
        html_str = Display._getMessageHtmlStr(html_msg, display_handler_name, **options)

        if html_str is not None:
            Display.show_html(html_str, display_handler_name=display_handler_name, **options)


    @staticmethod
    def _getInfoMessageHtmlStr(msg, display_handler_name=None, **options):
        return Display._getMessageHtmlStr(Display.getInfoMessageHtml(msg), display_handler_name=display_handler_name, **options)


    @staticmethod
    def showSuccessMessage(msg, display_handler_name=None, **options):
        Display._showMessage(Display.getSuccessMessageHtml(msg), display_handler_name=display_handler_name, **options)


    @staticmethod
    def showInfoMessage(msg, display_handler_name=None, **options):
        Display._showMessage(Display.getInfoMessageHtml(msg), display_handler_name=display_handler_name, **options)


    @staticmethod
    def showWarningMessage(msg, display_handler_name=None, **options):
        Display._showMessage(Display.getWarningMessageHtml(msg), display_handler_name=display_handler_name, **options)


    @staticmethod
    def showDangerMessage(msg, display_handler_name=None, **options):
        Display._showMessage(Display.getDangerMessageHtml(msg))



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import colorsys
from itertools import cycle
import re
from typing import List


from .dependencies import Dependencies
from .my_utils import is_collection


class Color(object):

    def __init__(self, rgb_color, name=None, **kwargs):
        self.color = rgb_color
        self.name = name or rgb_color


    def _repr_html_(self):
        return self._to_html()


    def _to_html(self):
        c = f'<div style="background-color:{self.color};height:20px;width:20px;display:inline-block;"></div>'
        return f'<div style="display:inline-block;padding:10px;"><div>{self.name}</div>{c}</div>'


    def __repr__(self):
        return self.color


class Palette(list):

    def __init__(self, palette_name=None, n_colors=None, desaturation=None, rgb_palette=None, range_start=None, to_reverse=False, **kwargs):
        self.name = palette_name or Palettes.get_default_pallete_name()
        self.n_colors = (n_colors or Palettes.DEFAULT_N_COLORS) if rgb_palette is None else len(rgb_palette)
        self.desaturation = desaturation or Palettes.DEFAULT_DESATURATION
        self.kwargs = kwargs
        self.range_start = range_start

        parsed = self.parse(self.name)
        self.name = parsed.get("name") or self.name

        if rgb_palette is None:
            rgb_palette = parsed.get("rgb_palette")
            if rgb_palette is None:
                rgb_float_pallete = self._get_color_palette(name=parsed.get("base_name"), n_colors=self.n_colors, desaturation=self.desaturation)
                rgb_palette = ["rgb" + str((int(rgb[0] * 255), int(rgb[1] * 255), int(rgb[2] * 255))) for rgb in rgb_float_pallete]
            if parsed.get("slice"):
                rgb_palette = rgb_palette.__getitem__(parsed.get("slice"))
            if parsed.get("reversed") is not None and parsed.get("reversed"):
                rgb_palette = list(reversed(rgb_palette))
            if to_reverse:
                rgb_palette = list(reversed(rgb_palette))
        super(Palette, self).__init__()
        self.extend(rgb_palette)


    def _to_html(self, add_details_to_name=True):
        name = self.name
        if self.range_start is not None:
            name = f"{name}[{self.range_start}:{self.range_start + len(self)}]"
        if add_details_to_name:
            desaturation_details = ""
            if self.desaturation is not None and self.desaturation > 0 and self.desaturation < 1.0:
                desaturation_details = f", desaturation {self.desaturation}"
            name = f"{name} ({self.n_colors} colors{desaturation_details})"
        s_s = ""
        for color in self:
            s_s = f'{s_s}<div style="background-color:{color};height:20px;width:20px;display:inline-block;"></div>'
            
        return f'<div style="display:inline-block;padding:10px;"><div>{name}</div>{s_s}</div>'


    def __getitem__(self, key):
        item = super(Palette, self).__getitem__(key)
        if isinstance(key, slice):
            range_start = min((key.start or 0), len(self)) + (self.range_start or 0)
            return Palette(palette_name=self.name, desaturation=self.desaturation, rgb_palette=item, range_start=range_start, **self.kwargs)
        else:
            return Color(item, name=f"{self.name}[{(self.range_start or 0) + key}]")


    def _repr_html_(self):
        return self._to_html()


    @classmethod
    def parse(cls, name):
        name = name.strip()
        reverse = name.endswith("_r")
        base_name = name[:-2] if reverse else name

        rgb_palette = None
        range = None

        if base_name.endswith("]"):
            start = base_name.rfind("[")
            # slice
            if start > 0:
                se_parts = [value.strip() for value in base_name[start + 1: -1].split(":")]
                if len(se_parts) == 2:
                    try:
                        range = slice(*[int(value) if value else None for value in se_parts])
                        base_name = base_name[:start]
                    except:
                        pass

        # custom
        if is_collection(base_name, "["):
            rgb_palette = eval(base_name.lower().replace("-", "").replace("_", ""))
            if not isinstance(rgb_palette, list) or len(rgb_palette) == 0:
                raise ValueError("invlaid custom palette syntax, should be a comma separate list.'[\"rgb(r,g,b)\",...]'")

            for rgb in rgb_palette:
                if not isinstance(rgb, str) or not rgb.startswith("rgb"):
                    raise ValueError("invlaid custom palette syntax, each item must have a 'rgb' prefix.'[\"rgb(r,g,b)\",\"rgb(...)\",...]'")
                color_list = eval(rgb[3:])
                if len(color_list) != 3:
                    raise ValueError("invlaid custom palette syntax, each color must be composed of a list of 3 number: \"rgb(r,g,b)\"")

                for color in color_list:
                    if not isinstance(color, int) or color < 0 or color > 255:
                        raise ValueError("invlaid custom palette syntax, each basic color (r,g,b) must between 0 to 255")
            name = name.lower().replace("-", "").replace("_", "").replace(" ", "")
            base_name = None

        return {"name": name, "base_name": base_name, "rgb_palette": rgb_palette, "reversed": reverse, "slice": range}


    @classmethod
    def validate_palette_name(cls, name):
        parsed = cls.parse(name)
        if parsed.get("rgb_palette") is None and parsed.get("base_name") not in Palettes.get_all_pallete_names():
            raise ValueError(
                f"must be a known palette name or custom palette (see option -popup_palettes) , but a value of {name} was specified."
            )


    @classmethod
    def validate_palette_desaturation(cls, desaturation):
        if desaturation > 1 or desaturation < 0:
            raise ValueError(f"must be between 0 and 1, but a value of {str(desaturation)} was specified.")


    @classmethod
    def validate_palette_colors(cls, n_colors):
        if n_colors < 1:
            raise ValueError(f"must be greater or equal than 1, but a value of {str(n_colors)} was specified.")

    @classmethod
    def _get_color_palette(cls, name=None, n_colors=1, desaturation=1):
        if name in Palettes.DEFAULT_PALETTES:       
            default_palette = Palettes.DEFAULT_PALETTES[name]
            pal_cycle = cycle(default_palette)
            palette = [next(pal_cycle) for _ in range(n_colors)]
            rgb_float_pallete = list(map(cls._rrggbb_to_rgb, palette))
            
        else:
            mplcmap = Dependencies.get_module('matplotlib.cm')
            mplcol = Dependencies.get_module('matplotlib.colors')
            
            if name in Palettes.MATPLOTLIB_DISTINCT_PALETTES:
                num = Palettes.MATPLOTLIB_DISTINCT_PALETTES[name]
                vector = [idx / (num - 1) for idx in range(0,num)][:n_colors]
            else:
                num = int(n_colors) + 2
                vector = [idx / (num - 1) for idx in range(0,num)][1:-1]

            color_map = mplcmap.get_cmap(name)
            palette = map(tuple, color_map(vector)[:, :3])

            rgb_float_pallete = list(map(mplcol.colorConverter.to_rgb, palette))

        rgb_float_pallete_desaturated = cls._desaturate_palette(rgb_float_pallete, desaturation)
        return rgb_float_pallete_desaturated


    @classmethod
    def _rrggbb_to_rgb(cls, hex_color):
        """Convert color in hex format #rrggbb or #rgb to an RGB color."""
        if isinstance(hex_color, str):
            # hex color in #rrggbb format.
            match = re.match(r"\A#[a-fA-F0-9]{6}\Z", hex_color)
            if match:
                return (tuple(int(value, 16) / 255 for value in [hex_color[1:3], hex_color[3:5], hex_color[5:7]]))
            # hex color in #rgb format, shorthand for #rrggbb.
            match = re.match(r"\A#[a-fA-F0-9]{3}\Z", hex_color)
            if match:
                return (tuple(int(value, 16) / 255 for value in [hex_color[1] * 2, hex_color[2] * 2, hex_color[3] * 2]))
        return hex_color


    @classmethod
    def _desaturate_palette(cls, reg_float_palette, desaturation):
        if not 0 <= desaturation <= 1:
            return [*reg_float_palette]

        return [cls._desaturate_rgb(rgb, desaturation) for rgb in reg_float_palette]


    @classmethod
    def _desaturate_rgb(cls, rgb, desaturation):
        if not 0 <= desaturation <= 1:
            return [*rgb]

        hue, lightness, saturation = colorsys.rgb_to_hls(*rgb)
        saturation *= desaturation
        saturated_rgb = colorsys.hls_to_rgb(hue, lightness, saturation)
        return saturated_rgb


class Palettes(list):

            # mplcmap = Dependencies.get_module('matplotlib.cm')
            # mplcol = Dependencies.get_module('matplotlib.colors')
    DEFAULT_DESATURATION = 1.0
    DEFAULT_N_COLORS = 10
    DEFAULT_NAME = "tab10" # should be from BASE_PALETTE_NAMES
    DEFAULT_ALT_NAME = "pastel" # should be from DEFAULT_PALETTES


    # DEFAULT_PALETTES: old matplotlib default palette
    DEFAULT_PALETTES = dict(
        deep=["#4C72B0", "#DD8452", "#55A868", "#C44E52", "#8172B3",
              "#937860", "#DA8BC3", "#8C8C8C", "#CCB974", "#64B5CD"],
        deep6=["#4C72B0", "#55A868", "#C44E52",
               "#8172B3", "#CCB974", "#64B5CD"],
        muted=["#4878D0", "#EE854A", "#6ACC64", "#D65F5F", "#956CB4",
               "#8C613C", "#DC7EC0", "#797979", "#D5BB67", "#82C6E2"],
        muted6=["#4878D0", "#6ACC64", "#D65F5F",
                "#956CB4", "#D5BB67", "#82C6E2"],
        pastel=["#A1C9F4", "#FFB482", "#8DE5A1", "#FF9F9B", "#D0BBFF",
                "#DEBB9B", "#FAB0E4", "#CFCFCF", "#FFFEA3", "#B9F2F0"],
        pastel6=["#A1C9F4", "#8DE5A1", "#FF9F9B",
                 "#D0BBFF", "#FFFEA3", "#B9F2F0"],
        bright=["#023EFF", "#FF7C00", "#1AC938", "#E8000B", "#8B2BE2",
                "#9F4800", "#F14CC1", "#A3A3A3", "#FFC400", "#00D7FF"],
        bright6=["#023EFF", "#1AC938", "#E8000B",
                 "#8B2BE2", "#FFC400", "#00D7FF"],
        dark=["#001C7F", "#B1400D", "#12711C", "#8C0800", "#591E71",
              "#592F0D", "#A23582", "#3C3C3C", "#B8850A", "#006374"],
        dark6=["#001C7F", "#12711C", "#8C0800",
               "#591E71", "#B8850A", "#006374"],
        colorblind=["#0173B2", "#DE8F05", "#029E73", "#D55E00", "#CC78BC",
                    "#CA9161", "#FBAFE4", "#949494", "#ECE133", "#56B4E9"],
        colorblind6=["#0173B2", "#029E73", "#D55E00",
                     "#CC78BC", "#ECE133", "#56B4E9"]
    )

    # matplotlib clormap + DEFAULT_PALETTES
    BASE_PALETTE_NAMES = [
        "deep",
        "muted",
        "bright",
        "pastel",
        "dark",
        "colorblind",

        "Accent",
        "Blues",
        "BrBG",
        "BuGn",
        "BuPu",
        "CMRmap",
        "Dark2",
        "GnBu",
        "Greens",
        "Greys",
        "OrRd",
        "Oranges",
        "PRGn",
        "Paired",
        "Pastel1",
        "Pastel2",
        "PiYG",
        "PuBu",
        "PuBuGn",
        "PuOr",
        "PuRd",
        "Purples",
        "RdBu",
        "RdGy",
        "RdPu",
        "RdYlBu",
        "RdYlGn",
        "Reds",
        "Set1",
        "Set2",
        "Set3",
        "Spectral",
        "Wistia",
        "YlGn",
        "YlGnBu",
        "YlOrBr",
        "YlOrRd",
        "afmhot",
        "autumn",
        "binary",
        "bone",
        "brg",
        "bwr",
        "cividis",
        "cool",
        "coolwarm",
        "copper",
        "cubehelix",
        "flag",
        "gist_earth",
        "gist_gray",
        "gist_heat",
        "gist_ncar",
        "gist_rainbow",
        "gist_stern",
        "gist_yarg",
        "gnuplot",
        "gnuplot2",
        "gray",
        "hot",
        "hsv",
        "inferno",
        "magma",
        "nipy_spectral",
        "ocean",
        "pink",
        "plasma",
        "prism",
        "rainbow",
        "seismic",
        "spring",
        "summer",
        "tab10",
        "tab20",
        "tab20b",
        "tab20c",
        "terrain",
        "viridis",
        "winter",
    ]

    MATPLOTLIB_DISTINCT_PALETTES = {
        "tab10": 10, "tab20": 20, "tab20b": 20, "tab20c": 20,
        "Set1": 9, "Set2": 8, "Set3": 12,
        "Accent": 8, "Paired": 12,
        "Pastel1": 9, "Pastel2": 8, "Dark2": 8,
    }

    all_palette_names = []
    default_palette_name = None

    @classmethod
    def get_all_pallete_names(cls)->List[str]:
        if len(cls.all_palette_names) == 0:
            mplcmap = Dependencies.get_module('matplotlib.cm', dont_throw=True)
            mplcol = Dependencies.get_module('matplotlib.colors', dont_throw=True)
            if mplcmap and mplcol:
                cls.all_palette_names = cls.BASE_PALETTE_NAMES
            else:
                cls.all_palette_names = list(cls.DEFAULT_PALETTES.keys())
        return cls.all_palette_names


    @classmethod
    def get_default_pallete_name(cls)->str:
        if cls.default_palette_name is None:
            if cls.DEFAULT_NAME in cls.get_all_pallete_names():
                cls.default_palette_name = cls.DEFAULT_NAME
            else:
                cls.default_palette_name = cls.DEFAULT_ALT_NAME
        return cls.default_palette_name


    def __init__(self, n_colors=None, desaturation=None, palette_list=None, to_reverse=False, **kwargs):
        self.n_colors = n_colors or Palettes.DEFAULT_N_COLORS
        self.desaturation = desaturation or Palettes.DEFAULT_DESATURATION
        self.to_reverse = to_reverse
        self.kwargs = kwargs
        super(Palettes, self).__init__()
        self.extend(palette_list or Palettes.get_all_pallete_names())


    def __getitem__(self, key):
        if isinstance(key, str):
            key = self.index(key)
        item = super(Palettes, self).__getitem__(key)
        if isinstance(key, slice):
            return Palettes(palette_list=item, desaturation=self.desaturation, n_colors=self.n_colors, to_reverse=self.to_reverse, **self.kwargs)
        else:
            return Palette(palette_name=item, desaturation=self.desaturation, n_colors=self.n_colors, to_reverse=self.to_reverse, **self.kwargs)


    def _to_html(self):
        n_colors = self.n_colors
        desaturation = self.desaturation
        suffix = f" (desaturation {str(desaturation)})" if desaturation is not None and desaturation != 1.0 and desaturation != 0 else ""
        html_str = f'<div style="text-align:center"><h1>{n_colors} colors palettes{suffix}</h1></div>'
        for name in self:
            for suffix in [""]:  # ['', '_r']:
                s = Palette(palette_name=name + suffix, n_colors=n_colors, desaturation=desaturation, **self.kwargs)
                html_str += s._to_html(add_details_to_name=False)
        return html_str


    def _repr_html_(self):
        return self._to_html()


    # plotly support this css colors:

"""       - A named CSS color:
            aliceblue, antiquewhite, aqua, aquamarine, azure,
            beige, bisque, black, blanchedalmond, blue,
            blueviolet, brown, burlywood, cadetblue,
            chartreuse, chocolate, coral, cornflowerblue,
            cornsilk, crimson, cyan, darkblue, darkcyan,
            darkgoldenrod, darkgray, darkgrey, darkgreen,
            darkkhaki, darkmagenta, darkolivegreen, darkorange,
            darkorchid, darkred, darksalmon, darkseagreen,
            darkslateblue, darkslategray, darkslategrey,
            darkturquoise, darkviolet, deeppink, deepskyblue,
            dimgray, dimgrey, dodgerblue, firebrick,
            floralwhite, forestgreen, fuchsia, gainsboro,
            ghostwhite, gold, goldenrod, gray, grey, green,
            greenyellow, honeydew, hotpink, indianred, indigo,
            ivory, khaki, lavender, lavenderblush, lawngreen,
            lemonchiffon, lightblue, lightcoral, lightcyan,
            lightgoldenrodyellow, lightgray, lightgrey,
            lightgreen, lightpink, lightsalmon, lightseagreen,
            lightskyblue, lightslategray, lightslategrey,
            lightsteelblue, lightyellow, lime, limegreen,
            linen, magenta, maroon, mediumaquamarine,
            mediumblue, mediumorchid, mediumpurple,
            mediumseagreen, mediumslateblue, mediumspringgreen,
            mediumturquoise, mediumvioletred, midnightblue,
            mintcream, mistyrose, moccasin, navajowhite, navy,
            oldlace, olive, olivedrab, orange, orangered,
            orchid, palegoldenrod, palegreen, paleturquoise,
            palevioletred, papayawhip, peachpuff, peru, pink,
            plum, powderblue, purple, red, rosybrown,
            royalblue, saddlebrown, salmon, sandybrown,
            seagreen, seashell, sienna, silver, skyblue,
            slateblue, slategray, slategrey, snow, springgreen,
            steelblue, tan, teal, thistle, tomato, turquoise,
            violet, wheat, white, whitesmoke, yellow,
            yellowgreen
            """

# Have colormaps separated into categories:
# http://matplotlib.org/examples/color/colormaps_reference.html
# cmaps = [('Perceptually Uniform Sequential', [
#             'viridis', 'plasma', 'inferno', 'magma']),
#          ('Sequential', [
#             'Greys', 'Purples', 'Blues', 'Greens', 'Oranges', 'Reds',
#             'YlOrBr', 'YlOrRd', 'OrRd', 'PuRd', 'RdPu', 'BuPu',
#             'GnBu', 'PuBu', 'YlGnBu', 'PuBuGn', 'BuGn', 'YlGn']),
#          ('Sequential (2)', [
#             'binary', 'gist_yarg', 'gist_gray', 'gray', 'bone', 'pink',
#             'spring', 'summer', 'autumn', 'winter', 'cool', 'Wistia',
#             'hot', 'afmhot', 'gist_heat', 'copper']),
#          ('Diverging', [
#             'PiYG', 'PRGn', 'BrBG', 'PuOr', 'RdGy', 'RdBu',
#             'RdYlBu', 'RdYlGn', 'Spectral', 'coolwarm', 'bwr', 'seismic']),
#          ('Qualitative', [
#             'Pastel1', 'Pastel2', 'Paired', 'Accent',
#             'Dark2', 'Set1', 'Set2', 'Set3',
#             'tab10', 'tab20', 'tab20b', 'tab20c']),
#          ('Miscellaneous', [
#             'flag', 'prism', 'ocean', 'gist_earth', 'terrain', 'gist_stern',
#             'gnuplot', 'gnuplot2', 'CMRmap', 'cubehelix', 'brg', 'hsv',
#             'gist_rainbow', 'rainbow', 'jet', 'nipy_spectral', 'gist_ncar'])]



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from typing import Union
import sys
import os
from datetime import datetime
import time
from threading import Thread, Event
from _thread import interrupt_main
import logging
from logging import Logger
import errno


from flask import Flask, send_file, request, make_response, Response


MAGIC_CLASS_NAME_UPPER = "KQLMAGIC"


def _get_env_var(var_name:str)->str:
    value = os.getenv(var_name)
    if value:
        # value = value.strip().upper().replace("_", "").replace("-", "")
        if value.startswith(("'", '"')):
            value = value[1:-1].strip()
    return value


logger:Logger = None


def init_logger(kernel_id:str, log_level:str=None, log_file:str=None, log_file_prefix:str=None, log_file_mode:str=None, log_to_console:bool=None)->Logger:
    global logger
    # create logger
    logger = logging.getLogger("kqlmagic-srv")
    default_log_level = logging.DEBUG

    log_level = log_level or _get_env_var(f"{MAGIC_CLASS_NAME_UPPER}_LOG_LEVEL")
    log_file = log_file or _get_env_var(f"{MAGIC_CLASS_NAME_UPPER}_SRV_LOG_FILE")
    log_file_prefix = log_file_prefix or _get_env_var(f"{MAGIC_CLASS_NAME_UPPER}_LOG_FILE_PREFIX")
    log_file_mode = log_file_mode or _get_env_var(f"{MAGIC_CLASS_NAME_UPPER}_LOG_FILE_MODE")

    log_to_console_str = _get_env_var(f"{MAGIC_CLASS_NAME_UPPER}_SRV_LOG_TO_CONSOLE")
    log_to_console = (log_to_console is None) or ((log_to_console_str is not None) and log_to_console_str.lower() == "true")

    # create file log handler
    create_file_logger_error_message = None
    file_handler = None
    try:
        if log_level or log_file or log_file_mode or log_file_prefix:

            log_level = log_level or default_log_level
            log_file = log_file or f"{log_file_prefix or 'kqlmagic'}-srv-{kernel_id}.log"
            # handler's default mode is 'a' (append)
            log_file_mode = (log_file_mode or "w").lower()[:1]
            file_handler = logging.FileHandler(log_file, mode=log_file_mode)
            file_handler.setLevel(log_level)
            formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
            file_handler.setFormatter(formatter)
            logger.addHandler(file_handler)

            if log_file_mode == "a":
                logger.info("\n\n----------------------------------------------------------------------\n\n")

            now = datetime.now()
            logger.info(f"start date {now.isoformat()}")
            logger.info(f"logger level {log_level}\n")

    except Exception as e:
        create_file_logger_error_message = f"failed to create file log handler. log_level: {log_level}, log_file: {log_file}, " \
                                           f"log_file_prefix: {log_file_prefix}, log_file_mode: {log_file_mode}, error: {e}"

    if log_to_console:
        # create console log handler
        console_handler = logging.StreamHandler()
        console_handler.setLevel(log_level or default_log_level)
        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)
    elif file_handler is None:
        # create null log handler
        null_handler = logging.NullHandler()
        logger.addHandler(null_handler)

    if create_file_logger_error_message is not None:
        logger.error(f"{create_file_logger_error_message}, continue log to console only.")
    return logger


class ParentHeartbeatMonitor(Thread):
    """ A heartbeat daemon thread that terminates the program immediately
    when the parent process no longer exists.
    """

    def __init__(self)->None:
        super(ParentHeartbeatMonitor, self).__init__()
        self.daemon = True
        self._stop_event = Event()
        self._heartbeat_event = Event()


    def run(self)->None:
        # We cannot use os.waitpid because it works only for child processes.

        count = 60
        logger.info(f"ParentHeartbeatMonitor for parent, parent is running.")

        while True:
            try:
                if count == 0:
                    count = 60
                    if self.heartbeated():
                        self._heartbeat_event.clear()
                        logger.info(f"ParentHeartbeatMonitor for parent, parent is running.")
                    else:
                        logger.info(f"ParentHeartbeatMonitor stopped to receive heartbeat from parent, shutting down.")
                        os_exit(0)
                        break
                count -= 1

                time.sleep(1.0)

                if self.stopped():
                    logger.info(f"ParentHeartbeatMonitor was asked to stop, exit.")
                    break

            except Exception as ex:
                error_message = f"ParentHeartbeatMonitor failed. error: {ex}"
                logger.error(f"{error_message}, continue monitoring.")


    def stop(self)->None:
        self._stop_event.set()


    def stopped(self)->bool:
        return self._stop_event.is_set()


    def heartbeat(self)->None:
        self._heartbeat_event.set()


    def heartbeated(self)->bool:
        return self._heartbeat_event.is_set()


class ParentProcessIdValueMonitor(Thread):
    """ A Unix-specific daemon thread that terminates the program immediately
    when the parent process no longer exists.
    """

    def __init__(self, platform:str)->None:
        super(ParentProcessIdValueMonitor, self).__init__()
        if platform == 'win32':
            raise Exception("specified liveness_mode: 'ppid' is not supported by 'win32' platform")
        elif parent_id == 1:
            # PID 1 (init) is special and will never go away,
            # only be reassigned.
            # Parent polling doesn't work if ppid == 1 to start with.
            raise Exception("specified liveness_mode: 'ppid' cannot work with parent_id == 1")
        self.daemon = True


    def run(self)->None:
        # We cannot use os.waitpid because it works only for child processes.

        count = 1

        while True:
            try:
                if count == 0:
                    count = 60
                    logger.info(f"ParentProcessIdValueMonitor for ppid {os.getppid()}, parent is running.")
                count -= 1

                time.sleep(1.0)

                if os.getppid() == 1:
                    logger.info(f"Parent appears to have exited, shutting down.")
                    os_exit(0)
                    break

            except OSError as e:
                if e.errno != errno.EINTR:
                    error_message = f"ParentProcessIdValueMonitor failed. error: {e}"
                    logger.error(f"{error_message}, continue monitoring.")
                    # raise

            except Exception as ex:
                error_message = f"ParentProcessIdValueMonitor failed. error: {ex}"
                logger.error(f"{error_message}, continue monitoring.")


test_disabled_modules = [module.strip() for module in (_get_env_var(f"{MAGIC_CLASS_NAME_UPPER}_TEST_DISABLE_MODULES") or "").split(",")]


def _get_import_module(module_name:str, message:str=None, package_name:str=None, dont_throw:bool=False):
    try:
        import importlib
        module = importlib.import_module(module_name)
        if module_name in test_disabled_modules:
            message = f", {message}" if message else ""
            message = f"module was disabled by {MAGIC_CLASS_NAME_UPPER}_TEST_DISABLE_MODULES{message}"
            raise Exception(message)
        return module
    except:
        if dont_throw:
            pass
        else:
            message = f", {message}" if message else ""
            import_error_message = f"failed to import '{module_name}' from '{package_name or module_name}'{message}"
            raise NotImplementedError(import_error_message)


class ParentProcessStateMonitor(Thread):
    """ A Windows-specific daemon thread that terminates the program immediately
    when the parent process no longer exists.
    """


    def __init__(self, parent_id:int=None)->None:
        super(ParentProcessStateMonitor, self).__init__()

        self.daemon = True
        self.parent_id = parent_id
        self.parent_create_time = None
        self.psutil = _get_import_module("psutil", message="won't be able to auto exit after parent's process exit")


    def run(self)->None:
        """ Run the monitor loop. This method never returns.
        """

        count = 0
        # Listen forever.
        while True:
            try:
                if count == 0:
                    count = 60
                    logger.info(f"ParentProcessStateMonitor for ppid {self.parent_id}, parent is running.")
                count -= 1

                time.sleep(1.0)

                parent_proc_not_found = True
                for proc in self.psutil.process_iter():
                    if proc.pid == self.parent_id:
                        proc_create_time = proc.create_time()
                        self.parent_create_time = self.parent_create_time or proc_create_time
                        # check proc.create_time to make sure the parent_id is not resued
                        if self.parent_create_time == proc_create_time:
                            parent_proc_not_found = False
                            break

                if parent_proc_not_found:
                    logger.info("Parent appears to have exited, shutting down.")
                    os_exit(0)
                    break
                    
            except OSError as e:
                if e.errno != errno.EINTR:
                    error_message = f"ParentProcessStateMonitor failed. error: {e}"
                    logger.error(f"{error_message}, continue monitoring.")
                    # raise

            except Exception as ex:
                error_message = f"ParentProcessStateMonitor failed. error: {ex}"
                logger.error(f"{error_message}, continue monitoring.")



DEFAULT_PORT = "5000"
DEFAULT_HOST = "127.0.0.1"
DEFAULT_PROTOCOL = "http"


parent_monitor = None
parent_kernel_id = None
base_folder = None
os_exit_started = None
params = {}
folder_list = []


app = Flask("kqlmagic_temp_files_server")


@app.after_request
def after_request_func(response:Response)->Response:
    """disable server and client cache"""
    try:
        if request is not None:
            logger.debug(f"{request.method} {request.url} - {response.status_code}")

        response.headers["Cache-Control"] = "no-cache, no-store, must-revalidate, public, max-age=0"
        response.headers["Expires"] = 0
        response.headers["Pragma"] = "no-cache"


    except Exception as ex:
        error_message = f"after_request_func failed. error: {ex}"
        logger.error(f"{error_message}.")

    return response


@app.route('/files/<foldername>/<kernel_id>/<filename>')
def files_1(foldername:str, kernel_id:str, filename:str)->Response:
    """return content of filename as the response body."""
    return _files(foldername, kernel_id, filename)


@app.route('/files/<foldername0>/<foldername1>/<kernel_id>/<filename>')
def files_2(foldername0:str, foldername1:str, kernel_id:str, filename:str)->Response:
    """return content of filename as the response body."""
    foldername = f"{foldername0}/{foldername1}"
    return _files(foldername, kernel_id, filename)


@app.route('/files/<foldername0>/<foldername1>/<foldername2>/<kernel_id>/<filename>')
def files_3(foldername0:str, foldername1:str, foldername2:str, kernel_id:str, filename:str)->Response:
    """return content of filename as the response body."""
    foldername = f"{foldername0}/{foldername1}/{foldername2}"
    return _files(foldername, kernel_id, filename)


def _files(foldername:str, kernel_id:str, filename:str)->Response:
    """return content of filename as the response body."""
    try:
        _kernel_id = request.args.get("kernelid")

        if len(folder_list) == 0 or f"{foldername}/{kernel_id}" in folder_list:
            err_resp = check_path(foldername, kernel_id, filename)
            if err_resp is not None:
                return err_resp
            file_path = f"{base_folder}/{foldername}/{kernel_id}/{filename}"

            logger.debug(f"/files/{foldername}/{kernel_id}/{filename} - {file_path}")
            return send_file(file_path)

        else:
            error_message = f"folder {foldername} not in {folder_list}"
            logger.error(f"{error_message}.")
            return make_response(f"{error_message}, internal error", 404)

    except Exception as ex:
        error_message = f"/files/<foldername>/<kernel_id>/<filename> failed, error: {ex}"
        logger.error(f"{error_message}.")
        return make_response(f"{error_message}, internal error", 500)


@app.route('/folders/<foldername>')
def folders(foldername:str)->Response:
    """return content of filename as the response body."""
    try:
        _kernel_id = request.args.get("kernelid")
        _filename = request.args.get("filename")

        if len(folder_list) == 0 or f"{foldername}/{_kernel_id}" in folder_list:
            err_resp = check_path(foldername, _kernel_id, _filename)
            if err_resp is not None:
                return err_resp
            file_path = f"{base_folder}/{foldername}/{_kernel_id}/{_filename}"

            logger.debug(f"folders/{foldername} - {file_path}")
            return send_file(file_path)

        else:
            error_message = f"/folder {foldername} not in {folder_list}"
            logger.error(f"{error_message}.")
            return make_response(f"{error_message}, internal error", 404)

    except Exception as ex:
        error_message = f"/folders/<foldername> failed. error: {ex}."
        logger.error(f"{error_message}.")
        return make_response(f"{error_message}, internal error", 500)
    


@app.route('/ping')
def ping()->Union[str,Response]:
    """print 'pong' as the response body."""
    try:
        _kernel_id = request.args.get("kernelid")

        pong = f"pong kernelid: {_kernel_id}"
        logger.debug(f"/ping - {pong}")
        return pong

    except Exception as ex:
        error_message = f"/ping failed. error: {ex}"
        logger.error(f"{error_message}.")
        return make_response(f"{error_message}, internal error", 500)


@app.route('/heartbeat')
def heartbeat()->Union[str,Response]:
    """print 'heartbeat' as the response body."""
    try:
        _kernel_id = request.args.get("kernelid")

        if parent_monitor.__class__.__name__ == ParentHeartbeatMonitor:
            parent_monitor.heartbeat()

        heartbeat = f"heartbeat kernelid: {_kernel_id}"
        logger.debug(f"/heartbeat - {heartbeat}")
        return heartbeat

    except Exception as ex:
        error_message = f"/heartbeat failed. error: {ex}"
        logger.error(f"{error_message}.")
        return make_response(f"{error_message}, internal error", 500)


@app.route('/webbrowser')
def webbrowser()->str:
    """open web browser."""
    try:
        _encoded_url = request.args.get("url")
        _kernel_id = request.args.get("kernelid")

        import urllib.parse
        import webbrowser

        url = urllib.parse.unquote(_encoded_url)
        webbrowser.open(url, new=1, autoraise=True)
        logger.debug(f"/webbrowser - {url} - close after 1 sec")

    except Exception as ex:
        error_message = f"/webbrowser failed. error: {ex}"
        logger.error(f"{error_message}.")
        pass

    return """<!DOCTYPE html>
            <html><body>
            <script>
            window.close();
            //setTimeout(function(){ window.close(); }, 1*1000);
            </script>
            </body></html>"""


@app.route('/abort')
def abort()->str:
    """aborts the process"""
    try:
        _kernel_id = request.args.get("kernelid")
        logger.debug(f"/abort")

        os_exit(0)

    except Exception as ex:
        error_message = f"/abort failed. error: {ex}"
        logger.error(f"{error_message}.")
        raise ex

    return ''


def check_path(foldername:str, kernel_id:str, filename:str)->Response:
    err_resp = None
    try:
        error_message = None
        if not os.path.exists(f"{base_folder}"):
            error_message = f"Base folder '{base_folder}' not found"

        elif not os.path.exists(f"{base_folder}/{foldername}"):
            error_message = f"Folder {base_folder}/{foldername} not found"

        elif not os.path.exists(f"{base_folder}/{foldername}/{kernel_id}"):
            error_message = f"Folder {base_folder}/{foldername}/{kernel_id}  not found"

        elif not os.path.exists(f"{base_folder}/{foldername}/{kernel_id}/{filename}"):
            error_message = f"File {base_folder}/{foldername}/{kernel_id}/{filename} not found"

        elif f"/{foldername}/{kernel_id}/{filename}".replace("\\", "/").find(f"/{os.pardir}/") >= 0:
            error_message = f"File {base_folder}/{foldername}/{kernel_id}/{filename} contains not allowed /{os.pardir}/"

        if error_message is not None:
            err_resp = make_response(error_message, 404)
            logger.error(f"check_path failed, {error_message}. code: 404.")

    except Exception as ex:
        error_message = f"File {base_folder}/{foldername}/{kernel_id}/{filename}, error: {ex}"
        logger.error(f"check_path failed, {error_message}, code: 500")
        err_resp = make_response(f"{error_message}, internal error", 500)

    return err_resp


def init_parent_monitor(platform:str, parent_id_str:str, liveness_mode:str)->Union[ParentHeartbeatMonitor,ParentProcessStateMonitor,ParentProcessIdValueMonitor]:
    if parent_id_str is not None:
        try:
            parent_id = int(parent_id_str) 

            if liveness_mode == "heartbeat":
                return ParentHeartbeatMonitor()

            elif liveness_mode == "parent_process_id_value":
                return ParentProcessIdValueMonitor(platform)

            elif liveness_mode == "parent_process_state":
                return ParentProcessStateMonitor(parent_id)

            elif liveness_mode == "show":
                return

            elif liveness_mode == None:
                #
                # legacy
                #
                if platform == 'win32':
                    return ParentProcessStateMonitor(parent_id)
                else:
                    try:
                        return ParentProcessIdValueMonitor(platform)
                    except:
                        return ParentProcessStateMonitor(parent_id)
            else:
                raise Exception(f"unknown liveness mode: {liveness_mode}")


       
        except Exception as ex:
            error_message = f"init_parent_monitor got an error: {ex}, parent monitor (liveness_mode: {liveness_mode} disabled"
            logger.error(f"{error_message}.")


def os_exit(code:int)->None:
    global os_exit_started
    if os_exit_started is None:
        os_exit_started = True
        try:
            logger.info(f"OS_EXIT code: {code}!!!")
            to_clean = params.get("clean", None)
            if to_clean is not None and len(folder_list) > 0:
                for folder in folder_list:
                    folder_path = None
                    try:
                        folder_path = f"{base_folder}/{folder}"
                        logger.info(f"start to clean folder: {folder_path}")
                        for filename in os.listdir(folder_path):
                            file_path = None
                            try:
                                file_path = f"{base_folder}/{folder}/{filename}"
                                os.unlink(file_path)
                            except Exception as e:
                                logger.error(f"failed to clean file: {file_path}, reason: {e}")
                        os.rmdir(folder_path)
                        logger.info(f"clean folder: {folder_path} done")
                    except Exception as e:
                        logger.error(f"failed to clean folder: {folder_path}, reason: {e}")
        except Exception as e:
            logger.error(f"failed to fully clean, reason: {e}")
        logger.info(f"clean finished")
        # time.sleep(60.0)
        os._exit(code)


if __name__ == "__main__":
    try:
        key = None
        for arg in sys.argv[1:]:
            kv = arg.split("=")
            if arg.startswith('-') and len(kv) == 2:
                key = kv[0][1:]
                value = kv[1]
                params[key] = value
                key= None

            elif arg.startswith('-'):
                key = arg[1:]
                params[key] = True
            elif key is not None:
                params[key] = arg
                key = None

        parent_kernel_id = params.get("parent_kernel_id")  
        logger = init_logger(parent_kernel_id)
        logger.debug(f"__name__ argv: {sys.argv[1:]}")
        logger.debug(f"__name__ params: {params}")
        base_folder = params.get("base_folder")       
        if base_folder is not None:
            base_folder = base_folder[:-1] if base_folder.endswith("/") else base_folder
            folders_str = params.get("folders", "")
            folder_list = [folder[:-1] if folder.endswith('/') else folder for folder in folders_str.split(",")]
            port = params.get("port", DEFAULT_PORT)
            host = params.get("host", DEFAULT_HOST)
            parent_id = params.get("parent_id", None) 
            liveness_mode = params.get("liveness_mode")
            parent_monitor = init_parent_monitor(sys.platform, parent_id, liveness_mode)
            if parent_monitor is not None:
                logger.info(f"start parent_monitor for parent id: {parent_id}")
                parent_monitor.start()
            logger.info(f" * parent id: {parent_id}")
            logger.info(f" * Base folder: {base_folder}")
            logger.info(f" * Folder list: {folder_list}")
            logger.info(f" * ")
            
            logger.info(f" **** ")
            app.run(host=host, port=port, debug=False, use_reloader=False, use_debugger=False)

        else:
            error_message = f"__name__ failed. base_folder value is None"
            logger.error(f"{error_message}.")
            time.sleep(5 * 60 * 1.0)
            os._exit(1)

    except Exception as ex:
        error_message = f"__name__ failed. error: {ex}"
        # this print is not for debug
        print(f"{error_message}.")
        try:
            if logger is not None:
                logger.error(f"{error_message}.")
        except:
            pass

        time.sleep(5 * 60 * 1.0)
        os._exit(1)



#!/usr/bin/python
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

# Note: None as default is valid
# commands that have no parameters, typekey should not exit or set to None
# default max_params is 1 if type is not None otherwise 0
# default min_params is 1 if type is not None otherwise 0
COMMANDS_TABLE = {
    "version": {"flag": "version", "type": None},
    "banner": {"flag": "banner", "type": None},
    "usage": {"flag": "usage", "type": None},
    "submit": {"flag": "submit", "type": None, "cell": True},  # default
    "help": {"flag": "help", "type": "str", "default": "help"},
    "faq": {"flag": "faq", "type": None},
    "palette": {"flag": "palette", "type": None},
    "palettes": {"flag": "palettes", "type": None},
    # "config": {"flag": "config", "type": "str", "default": None},
    "config": {"flag": "config", "type": "not_quoted_str", "allow_none": True, "max_params": 1, "min_params": 0},
    "bugreport": {"flag": "bug_report", "type": None},
    "conn": {"flag": "conn", "type": "str", "default": None},
    # should be per connection
    "cache": {"flag": "cache", "type": "str", "allow_none": True},
    "cachecreate": {"flag": "cache_create", "type": "str", "allow_none": True},
    "cacheappend": {"flag": "cache_append", "type": "str", "allow_none": True},
    "cachecreateorappend": {"flag": "cache_create_or_append", "type": "str", "allow_none": True},
    "cacheremove": {"flag": "cache_remove", "type": "str", "allow_none": True},
    "cachelist": {"flag": "cache_list", "type": None},
    "cachestop": {"flag": "cache_stop", "type": None},

    "usecache": {"flag": "use_cache", "type": "str", "allow_none": True},
    "usecachestop": {"flag": "use_cache_stop", "type": "str", "allow_none": True},
    "schema": {"flag": "schema", "type": "str", "default": None},
    "clearssodb": {"flag": "clear_sso_db", "type": None},
    "py": {"flag": "python", "type": "not_quoted_str", "allow_none": True, "max_params": 2, "min_params": 2, "cell": True},
    "pyro": {"flag": "python", "type": "not_quoted_str", "allow_none": True, "max_params": 2, "min_params": 2, "cell": True},
    "pyrw": {"flag": "python", "type": "not_quoted_str", "allow_none": True, "max_params": 2, "min_params": 2, "cell": True},
    "activatekernel": {"flag": "activate_kernel", "type": None},
    "deactivatekernel": {"flag": "deactivate_kernel", "type": None},

    "linemagic": {"flag": "line_magic", "type": "not_quoted_str", "allow_none": True, "max_params": 2, "min_params": 2},
    "cellmagic": {"flag": "cell_magic", "type": "not_quoted_str", "allow_none": True, "max_params": 3, "min_params": 3, "cell": True},
}



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from typing import Type, List, Dict


from .engine import Engine 
from .kusto_engine import KustoEngine
from .aria_engine import AriaEngine
from .ai_engine import AppinsightsEngine
from .aimon_engine import AimonEngine
from .la_engine import LoganalyticsEngine
from .cache_engine import CacheEngine
from .constants import ConnStrKeys, Schema
from .exceptions import KqlEngineError


class ConnectionError(Exception):
    pass


class Connection(object):

    _current_engine:Engine = None
    _engine_by_id:Dict[str,Engine] = {}
    _id_by_name:Dict[str,str] = {}
    _last_current_by_engine_class:Dict[Type[Engine], Engine] = {}

    _ENGINE_CLASS_LIST:Type[Engine] = [KustoEngine, AriaEngine, AppinsightsEngine, AimonEngine, LoganalyticsEngine, CacheEngine]
    _ENGINE_CLASS_BY_SCHEMA:Dict[str,Type[Engine]] = {}
    _RESERVED_CLUSTER_NAMES:List[str] = []
    
    # n = e = None
    engine_class:Type[Engine] = None
    for engine_class in _ENGINE_CLASS_LIST:
        for schema_name in engine_class.get_alt_uri_schema_names():
            _ENGINE_CLASS_BY_SCHEMA[schema_name] = engine_class
        reserved = engine_class.get_reserved_cluster_name()
        if reserved is not None:
            _RESERVED_CLUSTER_NAMES.append(reserved)


    @classmethod
    def is_empty(cls)->bool:
        return len(cls._engine_by_id) == 0


    @classmethod
    def _create_engine(cls, conn_str:str, user_ns:dict, **options)->Engine:
        engine_class = cls._find_engine_class_by_conn_str(conn_str)
        # wasn't found in connection list, but maybe a kusto database connection
        if engine_class is None:
            if "@" in conn_str:
                engine_class = KustoEngine
            else:
                valid_prefixes_str = ", ".join([f"{s}://" for s in cls._ENGINE_CLASS_BY_SCHEMA.keys()])
                raise KqlEngineError(f"invalid connection_str, unknown <uri schema name>. valid uri schemas are: {valid_prefixes_str}")

        last_current_engine:Engine = cls._last_current_by_engine_class.get(engine_class.__name__)

        if engine_class not in [KustoEngine, AriaEngine]:
            engine = engine_class(conn_str, user_ns, current=last_current_engine, **options)

        else:
            if "://" in conn_str:
                if last_current_engine is not None:
                    last_cluster_friendly_name = last_current_engine.get_cluster_friendly_name()
                    last_current_engine = cls._engine_by_id.get(f"@{last_cluster_friendly_name}")
                cluster_engine = engine_class(conn_str, user_ns, current=last_current_engine, conn_class=Connection, **options)

                cluster_friendly_name = cluster_engine.get_cluster_friendly_name()

                cls._save_and_set_current_engine(cluster_engine, name=f"@{cluster_friendly_name}")
                database_name = cluster_engine.get_database_name()
                alias = cluster_engine.get_alias()

            elif "@" in conn_str:
                database_name, cluster_friendly_name = conn_str.split("@", 1)
                alias = None
                if len(database_name) < 1:
                    raise KqlEngineError(f"invalid connection_str, key {ConnStrKeys.DATABASE} cannot be empty")
            else:
                raise KqlEngineError(f"invalid connection_str: {conn_str}")
            engine = cls._create_kusto_database_engine(engine_class, database_name, cluster_friendly_name, alias, user_ns, **options)

        if options.get("use_cache") and engine_class != CacheEngine:
            engine = CacheEngine(engine, user_ns, last_current_engine, **options)
        return engine


    @classmethod
    def _find_engine_class_by_conn_str(cls, conn_str:str)->Type[Engine]:
        if conn_str is not None:
            parts = conn_str.split("://", 1)
            if len(parts) == 2:
                uri_schema = parts[0].lower().replace("_", "").replace("-", "")
                return cls._ENGINE_CLASS_BY_SCHEMA.get(uri_schema)


    @classmethod
    def _create_kusto_database_engine(cls, engine_class:Engine,  database_name:str, cluster_friendly_name:str, alias:str, user_ns:dict, **options)->KustoEngine:
        if engine_class._RESERVED_CLUSTER_NAME != cluster_friendly_name and cluster_friendly_name in cls._RESERVED_CLUSTER_NAMES:
            raise KqlEngineError(
                f'invalid connection_str, connection_str pattern "{(alias or database_name)}@{cluster_friendly_name}" '
                f'cannot be used for reserved "appinsights", "loganalytics" and "cache"'
            )
        cluster_engine = cls.get_engine_by_name(f"@{cluster_friendly_name}")
        if cluster_engine is None:
            raise KqlEngineError(
                f'invalid connection_str, connection_str pattern "{(alias or database_name)}@{cluster_friendly_name}" '
                f'can be used only after a previous connection was established to cluster "{cluster_friendly_name}"'
            )
        cluster_name = cluster_engine.get_cluster_name()
        conn_dict = { 
            ConnStrKeys.DATABASE: database_name,
            ConnStrKeys.CLUSTER: cluster_name,
            ConnStrKeys.ALIAS: alias,
            ConnStrKeys.CLUSTER_FRIENDLY_NAME: cluster_friendly_name
        }
        
        return KustoEngine(conn_dict, user_ns, **options, conn_class=Connection)

            
    @classmethod
    def _save_and_set_current_engine(cls, engine:Engine, name:str=None)->Engine:
        "Save engine if new. Set as current engine"

        id = engine.get_id()
        cls._current_engine = cls._engine_by_id.get(id) or engine
        cls._engine_by_id[id] = cls._current_engine

        name = name or engine.get_conn_name()
        cls._engine_by_id[name] = cls._current_engine

        cls._id_by_name[name] = id
        cls._id_by_name[engine.bind_url] = id
        return cls._current_engine


    @classmethod
    def get_engine_by_name(cls, name:str)->Engine:
        id = cls._id_by_name.get(name, name)
        return cls._engine_by_id.get(id)


    @classmethod
    def get_engine(cls, conn_str:str, user_ns:dict, **options)->Engine:
        "Sets the current engine based on connection string"
        "If engine can't be found it is created "
        if conn_str:
            # either exist or create a new one
            cls._current_engine = cls.get_engine_by_name(conn_str)
            if cls._current_engine is None:
                engine = cls._create_engine(conn_str, user_ns, **options)
                cls._save_and_set_current_engine(engine)

        if not cls._current_engine:
            raise ConnectionError("No _current connection set yet.")
        cls._last_current_by_engine_class[cls._current_engine.__class__.__name__] = cls._current_engine        
        return cls._current_engine


    @classmethod
    def _get_sorted_name_list(cls)->List[str]:
        "returns only alias@cluster list"
        return [name for name in sorted(cls._id_by_name) if name.find("@") > 0 and name.find("://") < 0]


    @classmethod
    def get_connection_list_by_schema(cls, uri_schema_name:str)->List[str]:
        "returns alias@cluster list that are filtered by schema name"
        uri_schema_name = uri_schema_name.lower()
        return [name for name in cls._get_sorted_name_list() if uri_schema_name in cls.get_engine_by_name(name).get_alt_uri_schema_names()]


    @classmethod
    def get_connection_list_formatted(cls)->List[str]:
        "returns only alias@cluster list, list is formatted, current connection is prefixed with an asterics"
        name_list = []
        for name in cls._get_sorted_name_list():
            asterics_current = '*' if cls.get_engine_by_name(name) == cls._current_engine else ' '
            name_list.append(f" {asterics_current} {name}")
        return name_list


    @classmethod
    def get_connections_info(cls, name_filter_list:List[str]=None)->Dict[str,dict]:
        "returns connection info details per connection name"
        name_list = cls._get_sorted_name_list()
        for name in (name_filter_list or []):
            if name not in name_list:
                raise ValueError(f"connection: '{name}' not found")
        return {name: cls.get_engine_by_name(name).get_details() for name in name_list if (name_filter_list is None or name in name_filter_list)}


    @classmethod
    def get_current_connection_formatted(cls)->str:
        "returns current connection name formated"
        return f" * {cls._current_engine.get_conn_name()}"


    @classmethod
    def get_current_connection_name(cls)->str:
        "returns currennection name"
        return cls._current_engine.get_conn_name() if cls._current_engine else None
